/*******************************************************************************
 * segment.cpp: A hacked version of sndrms.c (by ?) found on the old CCRMA NeXT
 * machines back in the mid-90s.  This will use a very simple RMS method to try
 * and detect onsets in a sound file (so no fancy pitch/autocorrelation
 * detection).  I'm pretty sure the accompanying (old) sndfile.cpp code can
 * only handle 16 bit sound files.
 *
 * All of the attack times (in seconds) will be written to the terminal
 * (stdout) unless you specify and output file as the third argument.  (Don't
 * use the shell's > command to write to a file unless you want the preliminary
 * sound info to be written too.)  There are opening and closing parentheses
 * surrounding the timing data so that a whole file can be slurped up in Lisp
 * (with slippery-chicken's read-from-file, for example).
 *
 * Author: Michael Edwards (m@michael-edwards.org)
 *
 * Date: some time back in the 90s
 *
 * Usage: segment in.snd 
 * optional command-line args (relying on my memory of a long time ago): 
 * time-interval (default 0.010000).  How often in secs to calculate the RMS
 * threshold (default 0.060000). The amplitude increase we need to see between
 *    successive RMS calculations before we signal an onset
 * outfile (default: print onset times in second to terminal).  Text file of
 *    onset times to write 
 ******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include "sndfile.h"
#ifdef NEXT
#include <sound/sound.h>
#endif


#define amp_to_db(amp) (20 * (log10(fabs(amp))))
#define make_amps(struc, val, scl) struc.raw = val; struc.scaled = val * scl;\
  struc.db = amp_to_db(struc.scaled)
#define STRINGLENGTH 128
#define BUFSIZE 16777216
#define MINDIFF (0.001)  // Amp of .001 is -60 db

struct amps { double raw, scaled, db; };
struct extreme { char info[STRINGLENGTH]; double rms, time1, time2; };

/* Globals to hold highest and lowest RMS sections in the file.  0 is 
 * the RMS of silence and 32767 is the RMS of a straight line at amp=1
 * so these are good starting  values for the two. */
static struct extreme Highest = { "", 0, 0.0, 0.0 };
static struct extreme Lowest = { "", 32767, 0.0, 0.0 };
static double Buf[BUFSIZE];
static long BufIndex = 0;
static const double INTERVAL = 0.01;
static const double THRESHOLD = 0.06;
//static const double SHORTEST = INTERVAL;

/* Function prototypes */
short *get_data(char *, int *, int *, double *, int *);
void calc(double, double, double, double, int, double, double, int);
void print_env(void);
void finalise_env(double*, int);
void extract_notes(FILE* fp, double threshold);
void error(const char *function_name, const char *format,...);
int get_data(int start, int stop, double* dur, double* env, int maxpoints);

/*****************************************************************************/


int main(int argc, char *argv[])
{
  double sum = 0.0, sumsq = 0.0, max, min;
  double complete_sum = 0.0, complete_sumsq = 0.0;
  double complete_max = 0.0, complete_min = 0.0;
  double interval_secs = INTERVAL;
  double threshold = THRESHOLD;
  double time1 = 0.0, time2, dur;
  int srate, scount, ch, interval_samples, next_print;
  register short sp;
  register int temp1, temp2;
  int i;
  FILE* fp = NULL;


  if (argc < 2 || argc > 6) {
    fprintf(stderr,
            "\nUsage: %s soundfile (only 16bit .wav .aiff .snd) \n"
            "       optional: analysis interval [default %.2fsecs] \n"
            "                 threshold         [default %.2f] \n"
            "                 outfile           [default stdout]\n\n",
            argv[0], INTERVAL, THRESHOLD);
    exit(1);
  }
  
  sndfile sf(argv[1]);
  srate = sf.GetSrate();
  scount = sf.GetNumFrames() * sf.GetChannels();
  dur = sf.GetDuration();
  ch = sf.GetChannels();

  printf("%s:\n srate: %d, dur: %f, channels: %d, bit-depth: %d\n",
         argv[1], srate, dur,ch, sf.GetBitDepth());

  // this returns a float (-1.0 <-> 1.0) but we're working in 16 bit...
  sp = short(sf.NextSample() * 32767.0f);

  if (argc > 4) {
    fp = fopen(argv[4], "w");
    if (fp == NULL) {
      error("main", "Cannot open \"%s\" for writing.\n", argv[4]);
      return 1;
    }
  }
  else fp = stdout;
  if (argc > 2)
    interval_secs = atof(argv[2]);
  if (argc > 3)
    threshold = atof(argv[3]);
  time2 = interval_secs;
  interval_samples = next_print = (int)(interval_secs * srate * ch);
  max = min = sp;

  /* Main loop through sound file: */
  for (i = 0; i < scount; i++) {
    temp1 = sp;
    // sp = sf.NextSample();
    sp = short(sf.NextSample() * 32767.0f);
    // printf("\n%f", sp);
    temp2 = temp1 * temp1;
    sum += temp1;
    complete_sum += temp1;
    sumsq += temp2;
    complete_sumsq += temp2;
    if (max < temp1)
      max = temp1;
    else if (min > temp1)
      min = temp1;
    if (i == next_print) {
      calc(sum, sumsq, max, min, interval_samples, time1,
           time2, 1);
      time1 = time2;
      time2 += interval_secs;
      if (complete_max < max)
        complete_max = max;
      if (complete_min > min)
        complete_min = min;
      next_print += interval_samples;
      sum = sumsq = max = min = 0.0;
    }
  }
  /* This time don't check for extremes as we're at the end of the 
   * sound and it could be a very low amplitude--what we want is the
   * lowest point in the main body of the sound. */
  if (sum > 0.0) 
    calc(sum, sumsq, max, min, interval_samples -
         (next_print - scount), time1, dur, 0);
  Buf[BufIndex-1] = 0.0;
  //finalise_env(Buf, BufIndex);
  //print_env();
  fprintf(fp, "\n(");
  extract_notes(fp, threshold);
  fprintf(fp, ")\n");
  return 0;
}


/*****************************************************************************/


void calc(double sum, double sumsq, double max, double min,
                    int scount, double time1, double time2, int do_extremes)
{
  double mean = sum / ((double)scount);
  double meansq = sumsq / ((double)scount);
  double var = meansq - mean * mean;
  double rms = sqrt(var);
  const double scaler = pow(2, -15);
  struct amps RMS, MAX, MIN;
  char buf[STRINGLENGTH];
  
  make_amps(RMS, rms, scaler);
  make_amps(MAX, max, scaler);
  make_amps(MIN, min, scaler);
  sprintf(buf, "%.10g (%.10g) (%.10g dB)\n", RMS.raw, RMS.scaled, RMS.db);
  // printf("\n%f: %s", time1, buf);
  if (do_extremes) {
    if (rms > Highest.rms) {
      strcpy(Highest.info, buf);
      Highest.rms = rms;
      Highest.time1 = time1;
      Highest.time2 = time2;
    }
    if (rms < Lowest.rms && rms != 0.0) {  /* Ignore silence */
      strcpy(Lowest.info, buf);
      Lowest.rms = rms;
      Lowest.time1 = time1;
      Lowest.time2 = time2;
    }
  }

  if (BufIndex >= BUFSIZE) {
    fprintf(stderr, "Too many results.\n");
    exit(1);
  }
  Buf[BufIndex++] = time2;
  Buf[BufIndex++] = RMS.scaled;

}


/*****************************************************************************/

void finalise_env(double* env, int length)
{
  int i;
  double last = 0.0;
  double diff;

  for (i = 1; i < length; i+=2) {
    if (env[i] <= MINDIFF)
      env[i] = 0.0;
    diff = fabs(env[i] - last);
    if (diff <= MINDIFF)
      env[i] = -1.0;
    else {
      env[i-2] = last;
      last = env[i];
    }
  }
}

/*****************************************************************************/

void extract_notes(FILE* fp, double threshold)
{
  const int maxnotes = 100000;
  const int maxpoints = 100000;
  int i;
  int start1 = -1;     // Where RMS starts increasing
  int stop = -1;       // Where it stops increasing
  int start2 = -1;     // Where it starts increasing again
  int* indices = new int[maxnotes];
  double* env = new double[maxpoints];
  double dur;
  int j = 0;

  if (fp == NULL)
    fp = stdout;

  for (i = 3; i < BufIndex; i += 2) {
    if (Buf[i] == -1.0)
      continue;
    if (Buf[i] > Buf[i - 2] && start1 == -1)

      // We got the first ramp up.
      start1 = i - 2;
    else if (Buf[i] > Buf[i - 2] && stop != -1)

      // We got the second ramp up.
      start2 = i - 2;

    else if (Buf[i] < Buf[i - 2] && stop == -1 && start1 != -1)

      // We got the ramp down.
      stop = i - 2;

    if (start1 != -1 && stop != -1 && start2 != -1) {

      if ((stop - start1) >= 2
          && (Buf[stop] - Buf[start1]) > threshold)
        indices[j++] = start1 - 1;
      if (j > maxnotes)
        error("extract_notes", "Too many notes.\n");
      start1 = -1;
      stop = -1;
      start2 = -1;
    }
  }
  indices[j] = BufIndex;
  //for (i = 0; i < j; ++i)
  //printf("%f\n", Buf[indices[i]]);

#ifdef PRINT_LISP_ENV
  fprintf(fp, "(");
  for (i = 0; i < j; ++i) {
    int num_points = get_data(indices[i], indices[i + 1], &dur, env,
                              sizeof env / sizeof(double));
    fprintf(fp, "(%f %f\n(", Buf[indices[i]], dur);
    for (int k = 0; k < num_points; k+=2) {
      fprintf(fp, "%f %f ", env[k], env[k+1]);
      if ((k % 6) == 0)
        fputc('\n', fp);
    }
    fprintf(fp, "))\n\n");
  }
  fprintf(fp, ")\n\n");
#else
  for (i = 0; i < j; ++i) {
    get_data(indices[i], indices[i + 1], &dur, env,
             maxpoints);
    fprintf(fp, "%f\n", Buf[indices[i]]);
  }
#endif

  delete[] indices;
  delete[] env;
}


/*****************************************************************************/

int get_data(int start, int stop, double* dur, double* env, int maxpoints)
{
  int points = 2;
  env[0] = 0.0;
  env[1] = 0.0;
  double diff = 0.0;
  double last = 0.0;
  double dis = 0.0;
  const double st = Buf[start];
  double end = -1.0;

  for (int i = start + 2; i < stop; i += 2) {
    dis = Buf[i + 1];
    diff = fabs(dis - last);
    if (dis > MINDIFF && diff > MINDIFF) {
      env[points] = Buf[i] - st;
      env[points + 1] = dis;
      points += 2;
      end = -1.0;
    }
    else if (dis < MINDIFF && end == -1.0) {
      end = Buf[i] - st;
    }
    if (points > maxpoints) {
      error("get_data", "Too many points in envelope.\n");
      return 0;
    }
    last = dis;
  }
  env[points - 2] = *dur = end == -1.0 ? Buf[stop - 2] - st : end;
  env[points - 1] = 0.0;
  return points;
}


/*****************************************************************************/


void print_env(void)
{
  int i;
    
  printf("(0.0 0.0\n");
  for (i = 0; i < BufIndex; i+=2) {
    if (Buf[i+1] == -1.0)
      continue;
    printf(" %f %f\n", Buf[i], Buf[i+1]);
  }
  printf(")\n");
}

/*****************************************************************************/


// Simple error handling function.  Forces exit of programme. Works exactly
// like any of the printf family, except that it prints an introductory
// message supplying the function (<function_name>) that signalled the error
// as well as the user supplied message with all the usual formatting
// characteristics. 

/* 
void error(const char *function_name, const char *format,...)
{
    va_list args;

    va_start(args, format);
    fprintf(stderr, "Error signalled by function %s():\n\r", function_name);
    vfprintf(stderr, format, args);
    fprintf(stderr, "\n\r");
    va_end(args);
    exit(1);
}
*/

//*****************************************************************************

// EOF 
